import argparse
import asyncio
import os
import re
import signal
import sys

from asyncio.subprocess import PIPE, Process
from typing import List

from colorama import just_fix_windows_console

from bbb_dl.version import __version__
from bbb_dl.utils import (
    formatSeconds,
    Log,
    Timer,
    PathTools as PT,
)

_is_windows = sys.platform == "win32"


class BatchProcessor:
    def __init__(
        self,
        dl_urls_file_path: str,
        bbb_dl_path: str,
        output_dir: str,
        verbose: bool,
        no_check_certificate: bool,
        encoder: str,
        audiocodec: str,
        skip_webcam: bool,
        skip_webcam_freeze_detection: bool,
        skip_annotations: bool,
        skip_cursor: bool,
        keep_tmp_files: bool,
        ffmpeg_location: str,
        working_dir: str,
        backup: bool,
        max_parallel_chromes: int,
        force_width: int,
        force_height: int,
    ):
        self.bbb_dl_path = bbb_dl_path
        option_list = []
        self.add_bool_option(option_list, '--skip-webcam', skip_webcam)
        self.add_bool_option(option_list, '--skip-webcam-freeze-detection', skip_webcam_freeze_detection)
        self.add_bool_option(option_list, '--skip-annotations', skip_annotations)
        self.add_bool_option(option_list, '--skip-cursor', skip_cursor)
        self.add_bool_option(option_list, '--backup', backup)
        self.add_bool_option(option_list, '--verbose', verbose)
        self.add_bool_option(option_list, '--no-check-certificate', no_check_certificate)
        self.add_bool_option(option_list, '--keep-tmp-files', keep_tmp_files)
        self.add_value_option(option_list, '--ffmpeg-location', ffmpeg_location)
        self.add_value_option(option_list, '--working-dir', working_dir)
        self.add_value_option(option_list, '--output-dir', output_dir)
        self.add_value_option(option_list, '--encoder', encoder)
        self.add_value_option(option_list, '--audiocodec', audiocodec)
        self.add_value_option(option_list, '--max-parallel-chromes', max_parallel_chromes)
        self.add_value_option(option_list, '--force-width', force_width)
        self.add_value_option(option_list, '--force-height', force_height)
        self.default_option_list = option_list
        self.dl_urls_file_path = dl_urls_file_path

        self.output_dir_path = self.get_output_dir(output_dir)
        self.current_process = None

    def get_output_dir(
        self,
        output_dir: str,
    ):
        return self.check_directory(output_dir, os.getcwd(), 'output', '--output-dir')

    def check_directory(self, path: str, default_path: str, file_type: str, option_name: str):
        if path is None:
            path = default_path
        else:
            path = PT.sanitize_path(path)

        path = PT.get_abs_path(path)
        try:
            PT.make_dirs(path)
        except (OSError, IOError) as err:
            Log.error(f'Error: Unable to create directory "{path}" for {file_type} files: {str(err)}')
            Log.warning(
                f'You can choose an alternative directory for the {file_type} files with the {option_name} option.'
            )
            exit(-2)

        if not os.access(path, os.R_OK) or not os.access(path, os.W_OK):
            Log.error(f'Error: Unable to read or write in the directory for {file_type} files {path}')
            Log.warning(
                f'You can choose an alternative directory for the {file_type} files with the {option_name} option.'
            )
            exit(-3)
        return path

    def add_value_option(self, option_list, option_name, option):
        if option is not None:
            if option_name not in option_list:
                option_list.append(option_name)
                option_list.append(str(option))

    def add_bool_option(self, option_list, option_name, option):
        if option:
            if option_name not in option_list:
                option_list.append(option_name)

    def add_url_to_file(self, url: str, file_name: str):
        file_path = PT.get_in_dir(self.output_dir_path, file_name)
        with open(file_path, mode='a+', encoding='utf-8') as fh:
            fh.write(f"{url}\n")

    def run(self):
        if not os.path.isfile(self.dl_urls_file_path):
            Log.error(f'Can not find URLs file: {self.dl_urls_file_path}')
            exit(-1)

        URL_List = []
        try:
            with open(self.dl_urls_file_path, mode='r', encoding='utf-8') as fh:
                URL_List = [line.strip() for line in fh.readlines()]
        except OSError as err:
            Log.error(f'Error: {str(err)}')
            exit(-1)

        try:
            for url in URL_List:
                successful = asyncio.run(self.execute_bbb_dl(url))
                if successful:
                    self.add_url_to_file(url, 'successful.txt')
                else:
                    self.add_url_to_file(url, 'failed.txt')
        except KeyboardInterrupt:
            if self.current_process is not None:
                self.terminate()

    async def readlines(self, stream: asyncio.StreamReader):
        pattern = re.compile(rb"[\r\n]+")

        data = bytearray()
        while not stream.at_eof():
            lines = pattern.split(data)
            data[:] = lines.pop(-1)

            for line in lines:
                yield line

            data.extend(await stream.read(1024))

    def build_arguments(self, dl_url: str) -> List[str]:
        arguments = [self.bbb_dl_path, dl_url]
        arguments.extend(self.default_option_list)
        return arguments

    def create_async_subprocess(self, *args, **kwargs) -> Process:
        if _is_windows:
            # https://docs.python.org/3/library/asyncio-subprocess.html#asyncio.asyncio.subprocess.Process.send_signal
            from subprocess import CREATE_NEW_PROCESS_GROUP

            kwargs["creationflags"] = CREATE_NEW_PROCESS_GROUP

        return asyncio.create_subprocess_exec(*args, **kwargs)

    async def execute_bbb_dl(self, url: str) -> bool:
        arguments = self.build_arguments(url)

        self.current_process = await self.create_async_subprocess(
            *arguments,
            stdout=PIPE,
        )

        await asyncio.wait(
            [
                asyncio.create_task(self.read_stdout()),
                asyncio.create_task(self.current_process.wait()),
            ]
        )

        if self.current_process.returncode == 0:
            Log.success('Completed successfully')
            self.current_process = None
            return True
        else:
            Log.error(f'Error: {self.current_process.returncode}')
            self.current_process = None
            return False

    async def read_stdout(self):
        async for line in self.readlines(self.current_process.stdout):
            print(line.decode("utf-8"))

    def terminate(self):
        if self.current_process is None:
            return

        sigterm = signal.SIGTERM
        if _is_windows:
            sigterm = signal.CTRL_BREAK_EVENT

        self.current_process.send_signal(sigterm)


def get_parser():
    """
    Creates a new argument parser.
    """
    parser = argparse.ArgumentParser(description=('Big Blue Button Batch Downloader'))

    parser.add_argument(
        'URLs',
        type=str,
        help='Path to a text file containing URLs of BBB lessons, one line per URL',
    )

    parser.add_argument(
        '-bp',
        '--bbb-dl-path',
        type=str,
        default='bbb-dl',
        help='Path to bbb-dl. Use it if bbb-dl is not in your system PATH',
    )

    parser.add_argument(
        '-sw',
        '--skip-webcam',
        action='store_true',
        help='Skip adding the webcam video as an overlay to the final videos.',
    )
    parser.add_argument(
        '-swfd',
        '--skip-webcam-freeze-detection',
        action='store_true',
        help='Skip detecting if the webcam video is completely empty.'
        + ' It is assumed the webcam recordings are not empty.',
    )

    parser.add_argument(
        '-sa',
        '--skip-annotations',
        action='store_true',
        help='Skip capturing the annotations of the professor',
    )

    parser.add_argument(
        '-sc',
        '--skip-cursor',
        action='store_true',
        help='Skip capturing the cursor of the professor',
    )

    parser.add_argument(
        '-bk',
        '--backup',
        action='store_true',
        help=(
            'Downloads all the content from the server and then stops. After using this option, you can run bbb-dl'
            + ' again to create the video based on the saved files'
        ),
    )
    parser.add_argument(
        '-kt',
        '--keep-tmp-files',
        action='store_true',
        help=(
            'Keep the temporary files after finish. In case of an error bbb-dl will reuse the already generated files'
        ),
    )

    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help=('Print more verbose debug information'),
    )

    parser.add_argument(
        '--ffmpeg-location',
        type=str,
        default=None,
        help=(
            'Optional path to the directory in that your installed ffmpeg executable is located'
            + ' (Use it if ffmpeg is not located in your system PATH)'
        ),
    )

    parser.add_argument(
        '-ncc',
        '--no-check-certificate',
        action='store_true',
        help=('Suppress HTTPS certificate validation'),
    )

    parser.add_argument(
        '--version',
        action='version',
        version='bbb-dl ' + __version__,
        help='Print program version and exit',
    )

    parser.add_argument(
        '--encoder',
        dest='encoder',
        type=str,
        default='libx264',
        help='Optional encoder to pass to ffmpeg (default libx264)',
    )
    parser.add_argument(
        '--audiocodec',
        dest='audiocodec',
        type=str,
        default='copy',
        help='Optional audiocodec to pass to ffmpeg (default copy the codec from the original source)',
    )

    parser.add_argument(
        '-od',
        '--output-dir',
        type=str,
        default=None,
        help='Optional output directory for final videos',
    )

    parser.add_argument(
        '-wd',
        '--working-dir',
        type=str,
        default=None,
        help='Optional output directory for all temporary directories/files',
    )

    parser.add_argument(
        '-mpc',
        '--max-parallel-chromes',
        type=int,
        default=10,
        help='Maximum number of chrome browser instances used to generate frames',
    )

    parser.add_argument(
        '-fw',
        '--force-width',
        type=int,
        default=None,
        help='Force width on final outputs',
    )

    parser.add_argument(
        '-fh',
        '--force-height',
        type=int,
        default=None,
        help='Force height on final outputs',
    )

    return parser


# --- called at the program invocation: -------------------------------------
def main(args=None):
    just_fix_windows_console()
    parser = get_parser()
    args = parser.parse_args(args)

    with Timer() as final_t:
        BatchProcessor(
            args.URLs,
            args.bbb_dl_path,
            args.output_dir,
            args.verbose,
            args.no_check_certificate,
            args.encoder,
            args.audiocodec,
            args.skip_webcam,
            args.skip_webcam_freeze_detection,
            args.skip_annotations,
            args.skip_cursor,
            args.keep_tmp_files,
            args.ffmpeg_location,
            args.working_dir,
            args.backup,
            args.max_parallel_chromes,
            args.force_width,
            args.force_height,
        ).run()
    Log.info(f'BBB-DL finished and took: {formatSeconds(final_t.duration)}')



# Python script that downloads a lessen video from a published bbb recording.

import argparse
import asyncio
import hashlib
import math
import os
import re
import shutil
import traceback

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import partial
from http.server import ThreadingHTTPServer
from itertools import cycle
from pathlib import Path
from threading import Thread
from typing import List, Dict, Any, Tuple
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError, Element

import aiohttp
import aiofiles

from aiohttp.client_exceptions import ClientError, ClientResponseError
from playwright.async_api import async_playwright
from playwright.async_api._generated import Page
from colorama import just_fix_windows_console

from bbb_dl.ffmpeg import FFMPEG
from bbb_dl.utils import (
    _s,
    _x,
    append_get_idx,
    format_bytes,
    formatSeconds,
    get_free_port,
    Log,
    PathTools as PT,
    QuietRequestHandler,
    Timer,
    xpath_text,
)
from bbb_dl.version import __version__


class ActionType(Enum):
    show_image = 1
    hide_image = 2
    show_drawing = 3
    hide_drawing = 4
    set_view_box = 5
    move_cursor = 6


@dataclass
class Action:
    action_type: ActionType
    element_id: str = None
    value: Any = None
    width: int = None
    height: int = None
    x: int = None
    y: int = None


@dataclass
class Metadata:
    date: int
    date_formatted: str
    duration: float
    title: str
    bbb_version: str = None


@dataclass
class Frame:
    timestamp: float
    actions: [Action]
    capture_filename: str = None
    capture_path: str = None


@dataclass
class Deskshare:
    start_timestamp: float
    stop_timestamp: float
    width: int
    height: int


class BBBDL:
    VALID_URL_RE = re.compile(
        r'''(?x)
            (?P<website>https?://[^/]+)/playback/presentation/
            (?P<version>[\d\.]+)/
            (playback.html\?.*?meetingId=)?
            (?P<id>[0-9a-f\-]+)
        '''
    )
    NUMBER_RE = re.compile(r'\d+')

    headers = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Language': 'en',
        'Accept-Encoding': 'deflate, gzip',
    }

    def __init__(
        self,
        dl_url: str,
        filename: str,
        output_dir: str,
        verbose: bool,
        no_check_certificate: bool,
        encoder: str,
        audiocodec: str,
        skip_webcam: bool,
        skip_webcam_freeze_detection: bool,
        skip_annotations: bool,
        skip_cursor: bool,
        keep_tmp_files: bool,
        ffmpeg_location: str,
        working_dir: str,
        backup: bool,
        max_parallel_chromes: int,
        force_width: int,
        force_height: int,
    ):
        # Rendering options
        self.skip_webcam_opt = skip_webcam
        self.skip_webcam_freeze_detection_opt = skip_webcam_freeze_detection
        self.skip_annotations_opt = skip_annotations
        self.skip_cursor_opt = skip_cursor
        # BBB-dl Options
        self.keep_tmp_files = keep_tmp_files
        self.backup = backup
        self.working_dir = self.get_working_dir(working_dir)
        self.verbose = verbose
        self.no_check_certificate = no_check_certificate
        self.max_dl_retries = 10
        self.max_parallel_dl = 5
        self.max_parallel_chromes = int(max_parallel_chromes)

        # Job Options
        self.dl_url = dl_url
        self.filename = filename
        self.output_dir = self.get_output_dir(output_dir)
        self.force_width = int(force_width) if force_width is not None else None
        self.force_height = int(force_height) if force_height is not None else None

        self.ffmpeg = FFMPEG(verbose, ffmpeg_location, encoder, audiocodec)

        # Check DL-URL
        m_obj = re.match(self.VALID_URL_RE, self.dl_url)

        if m_obj is None:
            Log.error(
                f'Error: Your URL {self.dl_url} does not match the bbb session pattern.'
                + ' If you think this URL should work, please open an issue on https://github.com/C0D3D3V/bbb-dl/issues'
            )
            exit(-4)

        self.video_id = m_obj.group('id')
        self.video_website = m_obj.group('website')
        self.presentation_base_url = self.video_website + '/presentation/' + self.video_id
        self.tmp_dir = self.get_tmp_dir(self.video_id)
        self.frames_dir = self.get_frames_dir()

    def run(self):
        if not self.backup:
            Log.yellow(f'Output directory for the final video is: {self.output_dir}')
            Log.yellow(f'Directory for the temporary files is: {self.tmp_dir}')
        else:
            Log.yellow(f'Output directory for backup is: {self.tmp_dir}')

        Log.info("Downloading meta information")

        dl_jobs = ['metadata.xml', 'shapes.svg']
        _ = asyncio.run(self.batch_download_from_bbb(dl_jobs))

        Log.info("Downloading webcams / deskshare")
        dl_jobs = ['cursor.xml', 'panzooms.xml', 'captions.json', 'deskshare.xml', 'events.xml']
        cam_webm_idx = append_get_idx(dl_jobs, 'video/webcams.webm')
        cam_mp4_idx = append_get_idx(dl_jobs, 'video/webcams.mp4')
        dsk_webm_idx = append_get_idx(dl_jobs, 'deskshare/deskshare.webm')
        dsk_mp4_idx = append_get_idx(dl_jobs, 'deskshare/deskshare.mp4')

        dl_results = asyncio.run(self.batch_download_from_bbb(dl_jobs, False))

        if not dl_results[cam_webm_idx] and not dl_results[cam_mp4_idx]:
            Log.error('Error: webcams video is essential. Abort! Please try again later!')
            exit(4)
        webcams_rel_path = 'video/webcams.webm' if dl_results[cam_webm_idx] else 'video/webcams.mp4'
        webcams_path = PT.get_in_dir(self.tmp_dir, webcams_rel_path)

        deskshare_rel_path = (
            'deskshare/deskshare.webm'
            if dl_results[dsk_webm_idx]
            else 'deskshare/deskshare.mp4'
            if dl_results[dsk_mp4_idx]
            else None
        )
        deskshare_path = PT.get_in_dir(self.tmp_dir, deskshare_rel_path) if deskshare_rel_path is not None else None

        Log.info("Downloading slides")
        loaded_shapes = self.load_xml('shapes.svg')
        dl_jobs = self.get_all_image_urls(loaded_shapes)
        _ = asyncio.run(self.batch_download_from_bbb(dl_jobs))

        metadata = self.parse_metadata()
        deskshare_events = self.parse_deskshare_data(metadata.duration)
        if deskshare_path is None and len(deskshare_events) == 0:
            Log.yellow('No desk was shared in this session')
        elif deskshare_path is None and len(deskshare_events) > 0:
            Log.error(
                'Error: deskshare video is essential, because a desk was shared in this session.'
                + ' Abort! Please try again later!'
            )
            exit(5)

        if self.backup:
            Log.success("Backup Finished")
            Log.info("You can run bbb-dl again to generate the video based on the backed up files!")
            Log.yellow(f"Backup is located in: {self.tmp_dir}")
            return

        frames, only_zooms, partitions = self.parse_slides_data(loaded_shapes, metadata)
        self.create_frames(frames, only_zooms, partitions)

        slideshow_width, slideshow_height = self.get_slideshow_size(only_zooms, deskshare_path)
        if self.force_width is not None:
            slideshow_width = self.force_width
        if self.force_height is not None:
            slideshow_height = self.force_height
        slideshow_path = self.create_slideshow(frames, slideshow_width, slideshow_height)
        slideshow_path = self.add_deskshare_to_slideshow(
            slideshow_path, deskshare_path, deskshare_events, slideshow_width, slideshow_height, metadata
        )

        result_path = self.final_mux(
            slideshow_path, webcams_path, webcams_rel_path, slideshow_width, slideshow_height, metadata
        )

        if not self.keep_tmp_files:
            self.remove_tmp_dir()
        else:
            Log.warning(f'Temporary directory will not be deleted: {self.tmp_dir}')
        Log.success(f'All done! Final video: {result_path}')

    def parse_deskshare_data(self, recording_duration) -> List[Deskshare]:
        result_list = []
        loaded_deskshare = self.load_xml('deskshare.xml', False)
        if loaded_deskshare is None:
            return result_list
        deskshares = loaded_deskshare.findall("./event[@start_timestamp]")
        for deskshare in deskshares:
            deskshare_in = float(deskshare.get('start_timestamp'))
            deskshare_out = float(deskshare.get('stop_timestamp'))
            deskshare_width = int(deskshare.get('video_width'))
            deskshare_height = int(deskshare.get('video_height'))
            if deskshare_in < recording_duration:
                result_list.append(
                    Deskshare(
                        start_timestamp=deskshare_in,
                        stop_timestamp=deskshare_out,
                        width=deskshare_width,
                        height=deskshare_height,
                    )
                )
        result_list = sorted(result_list, key=lambda item: item.start_timestamp)
        return result_list

    def get_slideshow_size(self, only_zooms: Dict[float, Frame], deskshare_path: str):
        widths = []
        heights = []
        if deskshare_path is not None:
            video_info = self.ffmpeg.get_video_infos(deskshare_path)
            widths.append(video_info.width)
            heights.append(video_info.height)

        for _, frame in only_zooms.items():
            action = frame.actions[0]
            widths.append(int(action.width))
            heights.append(int(action.height))

        if len(widths) == 0 or len(heights) == 0:
            return

        max_width = max(widths)
        max_height = max(heights)

        if max_width % 2:
            max_width += 1
        if max_height % 2:
            max_height += 1

        return max_width, max_height

    def create_frames(self, frames: Dict[float, Frame], only_zooms: Dict[float, Frame], partitions: List[Tuple]):
        Log.info('Start capturing frames...')
        Log.info(f'Output directory for frames is: {self.frames_dir}')
        Log.info('Initialization takes a few seconds...')
        # Setup a server for Chrome browser to access
        port, port_error = get_free_port()
        if port is None:
            Log.error(f'Error: Could not open a port for a local http server: {port_error}')
            Log.warning(
                'Please check your Antivirus, to allow bbb-dl to open a local port.'
                + ' This is needed so we can use chrome browser to generate the presentation frames.'
            )
            exit(3)

        simple_handler = partial(QuietRequestHandler, directory=self.tmp_dir)
        server = ThreadingHTTPServer(('127.0.0.1', port), simple_handler)
        thread = Thread(target=server.serve_forever, daemon=True)
        thread.start()

        with Timer() as t:
            _ = asyncio.run(self.multi_capture_frames(f'http://localhost:{port}', frames, only_zooms, partitions))

        print()
        Log.info(f'Frames capturing is finished and took: {formatSeconds(t.duration)}.')

        server.shutdown()
        thread.join(timeout=10)

    async def display_capture_status(self, status_dict: Dict):
        spinner = cycle('/|\\-')
        print()
        while (status_dict.get('done', 0)) < status_dict.get('total', 0):
            print(
                "\r\033[KDone:"
                + f" {status_dict.get('done', 0):05} / {status_dict.get('total', 0):05} Frames"
                + f" | {status_dict.get('done_partitions', 0):03} / {status_dict.get('total_partitions', 0):03} Parts"
                + f" {next(spinner)}",
                end='',
            )
            await asyncio.sleep(1)

    async def _real_multi_capture_frames(
        self,
        server_url: str,
        frames: Dict[float, Frame],
        only_zooms: Dict[float, Frame],
        partitions: List[Tuple],
        status_dict: Dict,
    ):
        semaphore = asyncio.Semaphore(self.max_parallel_chromes)
        try:
            await asyncio.gather(
                *[
                    self.capture_frames(server_url, frames, only_zooms, partition, semaphore, status_dict)
                    for partition in partitions
                ]
            )
        except Exception:
            traceback.print_exc()
            Log.error(
                'Unexpected Error! Press Ctr+C to exit.'
                + ' Please try to set a low number of threads with `--max-parallel-chromes`.'
                + ' You can contact bbb-dl support.'
            )
            exit(-1)

    async def multi_capture_frames(
        self,
        server_url: str,
        frames: Dict[float, Frame],
        only_zooms: Dict[float, Frame],
        partitions: List[Tuple],
    ):
        status_dict = {
            'done': 0,
            'total': len(frames),
            'done_partitions': 0,
            'total_partitions': len(partitions),
        }
        await asyncio.wait(
            [
                asyncio.create_task(
                    self._real_multi_capture_frames(server_url, frames, only_zooms, partitions, status_dict)
                ),
                asyncio.create_task(self.display_capture_status(status_dict)),
            ],
        )

    async def capture_frames(
        self,
        server_url: str,
        frames: Dict[float, Frame],
        only_zooms: Dict[float, Frame],
        partition: Tuple,
        semaphore: asyncio.Semaphore,
        status_dict: Dict,
    ):
        async with semaphore, async_playwright() as p:
            first_timestamp = partition[0]
            last_timestamp = partition[1]
            partition_already_done = True
            total_frames_in_partition = 0
            for timestamp, frame in frames.items():
                if timestamp > last_timestamp:
                    break
                if timestamp < first_timestamp:
                    continue
                if not os.path.isfile(frame.capture_path):
                    partition_already_done = False
                    break
                total_frames_in_partition += 1

            if partition_already_done:
                status_dict['done'] += total_frames_in_partition
                print()
                Log.info(f'Partition already finished: {formatSeconds(partition[0])} to {formatSeconds(partition[1])}')
                status_dict['done_partitions'] += 1
                return

            browser = await p.chromium.launch()
            page = await browser.new_page()

            await page.goto(server_url + '/shapes.svg')
            await page.wait_for_selector('#svgfile')
            # add cursor
            await page.evaluate(
                """() => { 
                let el = document.querySelector('#svgfile')
                el.style.width = '100%'
                el.style.height = '100%'
                el.innerHTML = el.innerHTML + '<circle id="cursor" cx="9999" cy="9999" r="5" stroke="red" stroke-width="3" fill="red" style="visibility:hidden" />'
            }"""
            )
            current_view_box = None
            for timestamp, frame in only_zooms.items():
                if timestamp > first_timestamp:
                    continue
                # We only set one initial ViewBox, the first we find before the partition
                current_view_box = frame.actions[0]
                await self.set_view_box(page, current_view_box)
                break
            for timestamp, frame in frames.items():
                if timestamp > last_timestamp:
                    break
                if timestamp < first_timestamp:
                    continue
                for action in frame.actions:
                    if action.action_type == ActionType.show_image:
                        await self.show_image(page, action)
                        await self.show_cursor(page)
                    elif action.action_type == ActionType.hide_image:
                        await self.hide_image(page, action)
                        await self.hide_cursor(page)
                    elif action.action_type == ActionType.show_drawing:
                        await self.show_drawing(page, action)
                    elif action.action_type == ActionType.hide_drawing:
                        await self.hide_drawing(page, action)
                    elif action.action_type == ActionType.set_view_box:
                        current_view_box = action
                        await self.set_view_box(page, action)
                    elif action.action_type == ActionType.move_cursor:
                        if current_view_box is None:
                            Log.warning('No ViewBox, cursor position unclear!')
                            await self.move_cursor(page, -1, -1)
                        if current_view_box is not None:
                            if action.x == -1 and action.y == -1:
                                await self.move_cursor(page, -1, -1)
                            else:
                                await self.move_cursor(
                                    page,
                                    current_view_box.x + (action.x * current_view_box.width),
                                    current_view_box.y + (action.y * current_view_box.height),
                                )

                if not os.path.isfile(frame.capture_path):
                    await page.screenshot(path=frame.capture_path)
                status_dict['done'] += 1

            await browser.close()
            print()
            Log.info(f'Partition finished: {formatSeconds(partition[0])} to {formatSeconds(partition[1])}')
            status_dict['done_partitions'] += 1

    async def show_image(self, page: Page, action: Action):
        await page.evaluate(
            """([id, canvas_num]) => {
                document.querySelector('#' + id).style.visibility = 'visible'
                const canvas = document.querySelector('#canvas' + canvas_num)
                if (canvas) canvas.setAttribute('display', 'block')
            }""",
            [action.element_id, action.value],
        )

    async def hide_image(self, page: Page, action: Action):
        await page.evaluate(
            """([id, canvas_num]) => {
                document.querySelector('#' + id).style.visibility = 'hidden'
                const canvas = document.querySelector('#canvas' + canvas_num)
                if (canvas) canvas.setAttribute('display', 'none')
            }""",
            [action.element_id, action.value],
        )

    async def show_drawing(self, page: Page, action: Action):
        await page.evaluate(
            """([id, shape_id]) => {
                document.querySelectorAll('[shape=' + shape_id + ']').forEach( element => {
                    element.style.visibility = 'hidden'
                })
                document.querySelector('#' + id).style.visibility = 'visible'
            }""",
            [action.element_id, action.value],
        )

    async def hide_drawing(self, page: Page, action: Action):
        await page.evaluate(
            """(id) => {
                document.querySelector('#' + id).style.display = 'none'
            }""",
            action.element_id,
        )  # Maybe use visibility?

    async def set_view_box(self, page: Page, action: Action):
        await page.set_viewport_size({"width": int(action.width), "height": int(action.height)})
        await page.evaluate(
            """(viewBox) => {
                document.querySelector('#svgfile').setAttribute('viewBox', viewBox)
            }""",
            action.value,
        )

    async def show_cursor(self, page: Page):
        await page.evaluate(
            """() => {
                document.querySelector('#cursor').style.visibility = 'visible'
            }""",
        )

    async def hide_cursor(self, page: Page):
        await page.evaluate(
            """() => {
                document.querySelector('#cursor').style.visibility = 'hidden'
            }""",
        )

    async def move_cursor(self, page: Page, x: float, y: float):
        await page.evaluate(
            """([x,y]) => {
                document.querySelector('#cursor').setAttribute('cx', x)
                document.querySelector('#cursor').setAttribute('cy', y)
            }""",
            [x, y],
        )

    def get_all_image_urls(self, loaded_shapes: Element) -> List[str]:
        image_urls = []
        shapes_images = loaded_shapes.findall(_s(".//svg:image"))
        for image in shapes_images:
            image_rel_path = image.get(_x('xlink:href'))
            if image_rel_path not in image_urls:
                image_urls.append(image_rel_path)
        return image_urls

    def parse_metadata(self) -> Metadata:
        loaded_metadata = self.load_xml('metadata.xml')

        date = xpath_text(loaded_metadata, 'start_time')  # date on that the recording took place
        date_formatted = datetime.fromtimestamp(int(date) / 1000).strftime('%Y-%m-%dT%H-%M-%S')
        duration = float(xpath_text(loaded_metadata, './playback/duration')) / 1000.0  # in seconds
        title = xpath_text(loaded_metadata, './meta/meetingName')

        Log.info(f"Recording title: {title}")
        Log.info(f"Recording date: {date_formatted}")
        Log.info(f"Recording duration: {formatSeconds(duration)}")

        bbb_version = None
        if self.verbose:
            try:
                bbb_origin_version = xpath_text(loaded_metadata, './meta/bbb-origin-version')
                if bbb_origin_version is not None:
                    bbb_version = bbb_origin_version.split(' ')[0]
                    Log.info(f"BBB version: {bbb_version}")
            except IndexError:
                pass

        return Metadata(date, date_formatted, duration, title, bbb_version)

    def parse_slides_data(self, loaded_shapes: Element, metadata: Metadata) -> Dict[float, Frame]:
        frames = {}

        partitions = self.parse_slide_partitions(loaded_shapes, metadata.duration)
        self.parse_images(loaded_shapes, frames, metadata.duration)
        if not self.skip_annotations_opt:
            self.parse_drawings(loaded_shapes, frames, metadata.duration)

        only_zooms = {}
        loaded_zooms = self.load_xml('panzooms.xml', False)
        if loaded_zooms is not None:
            self.parse_zooms(loaded_zooms, frames, only_zooms, metadata.duration)

        if not self.skip_cursor_opt:
            loaded_cursors = self.load_xml('cursor.xml', False)
            if loaded_cursors is not None:
                self.parse_cursors(loaded_cursors, frames, metadata.duration)

        frames = dict(sorted(frames.items(), key=lambda item: item[0]))
        only_zooms = dict(sorted(only_zooms.items(), key=lambda item: item[0], reverse=True))

        return frames, only_zooms, partitions

    def get_frame_by_timestamp(self, frames: Dict[float, Frame], timestamp: float):
        if timestamp not in frames:
            capture_filename = f'{timestamp}.png'
            capture_path = PT.get_in_dir(self.frames_dir, capture_filename)
            frames[timestamp] = Frame(timestamp, [], capture_filename, capture_path)
        return frames[timestamp]

    def parse_slide_partitions(self, loaded_shapes: Element, recording_duration: float) -> List[Tuple]:
        partitions = []
        slides = loaded_shapes.findall(_s("./svg:image[@class='slide']"))
        partition_start = None
        last_slide_idx = len(slides) - 1
        for idx, image in enumerate(slides):
            image_id = image.get('id')
            image_in = float(image.get('in'))
            image_out = float(image.get('out'))
            got_annotations = loaded_shapes.find(_s(f"./svg:g[@image='{image_id}']")) is not None
            if partition_start is None:
                partition_start = image_in
            if idx == last_slide_idx or got_annotations:
                partitions.append((partition_start, image_out))
                partition_start = None
        return partitions

    def parse_images(self, loaded_shapes: Element, frames: Dict[float, Frame], recording_duration: float):
        slides = loaded_shapes.findall(_s("./svg:image[@class='slide']"))
        for image in slides:
            image_id = image.get('id')
            image_id_value = self.NUMBER_RE.search(image_id).group()
            image_in = float(image.get('in'))
            image_out = float(image.get('out'))
            image_width = int(float(image.get('width')))
            image_height = int(float(image.get('height')))
            if image_in < recording_duration:
                self.get_frame_by_timestamp(frames, image_in).actions.append(
                    Action(
                        action_type=ActionType.show_image,
                        element_id=image_id,
                        value=image_id_value,
                        width=image_width,
                        height=image_height,
                    )
                )
                self.get_frame_by_timestamp(frames, min(recording_duration, image_out)).actions.append(
                    Action(
                        action_type=ActionType.hide_image,
                        element_id=image_id,
                        value=image_id_value,
                    )
                )

    def parse_drawings(self, loaded_shapes: Element, frames: Dict[float, Frame], recording_duration: float):
        drawings = loaded_shapes.findall(_s(".//svg:g[@timestamp]"))
        for drawing in drawings:
            drawing_id = drawing.get('id')
            drawing_shape_value = drawing.get('shape')
            drawing_in = float(drawing.get('timestamp'))
            drawing_out = float(drawing.get('undo'))
            if drawing_in < recording_duration:
                self.get_frame_by_timestamp(frames, drawing_in).actions.append(
                    Action(
                        action_type=ActionType.show_drawing,
                        element_id=drawing_id,
                        value=drawing_shape_value,
                    )
                )
                if drawing_out != -1:
                    self.get_frame_by_timestamp(frames, min(recording_duration, drawing_out)).actions.append(
                        Action(
                            action_type=ActionType.hide_drawing,
                            element_id=drawing_id,
                        )
                    )

    def parse_zooms(
        self,
        loaded_zooms: Element,
        frames: Dict[float, Frame],
        only_zooms: Dict[float, Frame],
        recording_duration: float,
    ):
        zooms = loaded_zooms.findall("./event[@timestamp]")
        for zoom in zooms:
            zoom_in = float(zoom.get('timestamp'))
            zoom_value = zoom.find('viewBox').text
            zoom_value_split = zoom_value.split(' ')  # min-x min-y width height
            zoom_x = float(zoom_value_split[0])
            zoom_y = float(zoom_value_split[1])
            zoom_width = float(zoom_value_split[2])
            zoom_height = float(zoom_value_split[3])
            if zoom_in < recording_duration:
                zoom_action = Action(
                    action_type=ActionType.set_view_box,
                    value=zoom_value,
                    x=zoom_x,
                    y=zoom_y,
                    width=zoom_width,
                    height=zoom_height,
                )
                self.get_frame_by_timestamp(frames, zoom_in).actions.append(zoom_action)
                self.get_frame_by_timestamp(only_zooms, zoom_in).actions.append(zoom_action)

    def parse_cursors(self, loaded_cursors: Element, frames: Dict[float, Frame], recording_duration: float):
        cursors = loaded_cursors.findall("./event[@timestamp]")
        for cursor in cursors:
            cursor_in = float(cursor.get('timestamp'))
            cursor_value_text = cursor.find('cursor').text.split(' ')
            cursor_x = float(cursor_value_text[0])
            cursor_y = float(cursor_value_text[1])
            cursor_value = (float(cursor_value_text[0]), float(cursor_value_text[1]))
            if cursor_in < recording_duration:
                self.get_frame_by_timestamp(frames, cursor_in).actions.append(
                    Action(
                        action_type=ActionType.move_cursor,
                        x=cursor_x,
                        y=cursor_y,
                        value=cursor_value,
                    )
                )

    def get_output_file_path(self, metadata: Metadata):
        if self.filename is not None:
            return str(Path(self.output_dir) / PT.to_valid_name(self.filename))
        else:
            return str(
                Path(self.output_dir) / PT.to_valid_name(metadata.date_formatted + '_' + metadata.title + '.mp4')
            )

    def get_output_dir(
        self,
        output_dir: str,
    ):
        return self.check_directory(output_dir, os.getcwd(), 'output', '--output-dir')

    def get_working_dir(
        self,
        working_dir: str,
    ):
        return self.check_directory(working_dir, PT.get_project_data_directory(), 'temporary', '--working-dir')

    def check_directory(self, path: str, default_path: str, file_type: str, option_name: str):
        if path is None:
            path = default_path
        else:
            path = PT.sanitize_path(path)

        path = PT.get_abs_path(path)
        try:
            PT.make_dirs(path)
        except (OSError, IOError) as err:
            Log.error(f'Error: Unable to create directory "{path}" for {file_type} files: {str(err)}')
            Log.warning(
                f'You can choose an alternative directory for the {file_type} files with the {option_name} option.'
            )
            exit(-2)

        if not os.access(path, os.R_OK) or not os.access(path, os.W_OK):
            Log.error(f'Error: Unable to read or write in the directory for {file_type} files {path}')
            Log.warning(
                f'You can choose an alternative directory for the {file_type} files with the {option_name} option.'
            )
            exit(-3)
        return path

    def get_frames_dir(self):
        frames_dir = PT.get_in_dir(self.tmp_dir, 'frames')
        try:
            PT.make_dirs(frames_dir)
        except (OSError, IOError) as err:
            Log.error(f'Error: Unable to create directory "{frames_dir}" for generated frames: {str(err)}')
            exit(-7)

        return frames_dir

    def get_tmp_dir(self, video_id):
        # We use a shorted version of the video id as name for the temporary directory
        short_video_id = hashlib.md5(video_id.encode(encoding='utf-8')).hexdigest()

        tmp_dir = PT.get_in_dir(self.working_dir, short_video_id)
        try:
            PT.make_dirs(tmp_dir)
        except (OSError, IOError) as err:
            Log.error(f'Error: Unable to create directory "{tmp_dir}" for temporary files: {str(err)}')
            exit(-5)

        return tmp_dir

    def remove_tmp_dir(self):
        Log.info("Cleanup")
        try:
            if os.path.exists(self.tmp_dir):
                shutil.rmtree(self.tmp_dir)
        except (OSError, IOError) as err:
            Log.error(f'Error: Unable to remove directory "{self.tmp_dir}" for temporary files: {str(err)}')
            exit(-6)

    def get_bbb_link(self, rel_file_path: str):
        assert not rel_file_path.startswith('/') and not rel_file_path.startswith('\\')
        return self.presentation_base_url + '/' + rel_file_path

    async def readexactly(self, steam, n):
        if steam._exception is not None:
            raise steam._exception

        blocks = []
        while n > 0:
            block = await steam.read(n)
            if not block:
                break
            blocks.append(block)
            n -= len(block)

        return b''.join(blocks)

    async def get_can_continue_on_fail(self, url, session):
        try:
            headers = self.headers.copy()
            headers['Range'] = 'bytes=0-4'
            resp = await session.request("GET", url, headers=headers)
            return resp.headers.get('Content-Range') is not None and resp.status == 206
        except Exception as err:
            if self.verbose:
                Log.debug(f"Failed to check if download can be continued on fail: {err}")
        return False

    async def batch_download_from_bbb(self, dl_jobs: List[str], is_essential: bool = True) -> List[bool]:
        """
        @param dl_jobs: List of rel_file_path
        @param is_essential: Applied to all jobs
        """
        semaphore = asyncio.Semaphore(self.max_parallel_dl)
        dl_results = await asyncio.gather(
            *[self.download_from_bbb(dl_job, is_essential, semaphore) for dl_job in dl_jobs]
        )
        if is_essential:
            for idx, downloaded in enumerate(dl_results):
                if not downloaded:
                    Log.error(f'Error: {dl_jobs[idx]} is essential. Abort! Please try again later!')
                    exit(1)
        return dl_results

    async def download_from_bbb(
        self,
        rel_file_path: str,
        is_essential: bool,
        semaphore: asyncio.Semaphore,
        conn_timeout: int = 10,
        read_timeout: int = 1800,
    ) -> bool:
        """Returns True if the file was successfully downloaded or exists"""
        local_path = PT.get_in_dir(self.tmp_dir, rel_file_path)
        if os.path.exists(local_path):
            # Warning: We do not check if the file is complete
            Log.info(f'{rel_file_path} is already present')
            return True
        else:
            PT.make_base_dir(local_path)
            dl_url = self.get_bbb_link(rel_file_path)
            if self.verbose:
                Log.info(f'Downloading {rel_file_path} from: {dl_url}')
            else:
                Log.info(f'Downloading {rel_file_path}...')

            received = 0
            total = 0
            tries_num = 0
            file_obj = None
            can_continue_on_fail = False
            headers = self.headers.copy()
            finished_successfully = False
            async with semaphore, aiohttp.ClientSession(
                conn_timeout=conn_timeout, read_timeout=read_timeout
            ) as session:
                while tries_num < self.max_dl_retries:
                    try:
                        if tries_num > 0 and can_continue_on_fail:
                            headers["Range"] = f"bytes={received}-"
                        ssl_param = False if self.no_check_certificate else None
                        async with session.request(
                            "GET", dl_url, headers=headers, raise_for_status=True, ssl=ssl_param
                        ) as resp:

                            # Download the file.
                            total = int(resp.headers.get("Content-Length", 0))
                            content_range = resp.headers.get("Content-Range", "")  # Example: bytes 200-1000/67589

                            if resp.status not in [200, 206]:
                                if self.verbose:
                                    Log.debug(f"Warning {rel_file_path} got status {resp.status}")

                            if tries_num > 0 and can_continue_on_fail and not content_range and resp.status != 206:
                                raise ClientError(
                                    f"Server did not response for {rel_file_path} with requested range data"
                                )
                            file_obj = file_obj or await aiofiles.open(local_path, "wb")
                            chunk = await self.readexactly(resp.content, 1024000)
                            chunk_idx = 0
                            while chunk:
                                received += len(chunk)
                                if chunk_idx % 10 == 0:
                                    Log.info(f"{rel_file_path} got {format_bytes(received)} / {format_bytes(total)}")
                                await file_obj.write(chunk)
                                chunk = await self.readexactly(resp.content, 1024000)
                                chunk_idx += 1

                        if self.verbose:
                            Log.success(f'Downloaded {rel_file_path} to: {local_path}')
                        else:
                            Log.success(f'Successfully downloaded {rel_file_path}')

                        finished_successfully = True
                        break

                    except (ClientError, OSError, ValueError) as err:
                        if tries_num == 0:
                            can_continue_on_fail = await self.get_can_continue_on_fail(dl_url, session)
                        if not can_continue_on_fail:
                            # Clean up failed file because we can not recover
                            if file_obj is not None:
                                file_obj.close()
                            if os.path.exists(local_path):
                                os.unlink(local_path)

                        if isinstance(err, ClientResponseError) and err.status == 404 and not is_essential:
                            Log.info(f'{rel_file_path} could not be downloaded: {err.status} {err.message}')
                            if self.verbose:
                                Log.info(f'Error: {str(err)}')
                            break

                        if self.verbose:
                            Log.warning(
                                f'(Try {tries_num} of {self.max_dl_retries})'
                                + f' Unable to download "{rel_file_path}": {str(err)}'
                            )
                        tries_num += 1

            if file_obj is not None:
                file_obj.close()
            if not finished_successfully:
                return False
            return True

    def load_xml(self, rel_file_path: str, is_essential: bool = True):
        local_path = PT.get_in_dir(self.tmp_dir, rel_file_path)
        if os.path.exists(local_path):
            try:
                tree_root = ElementTree.parse(local_path).getroot()
                return tree_root
            except ParseError as err:
                Log.error(f'Unable to parse XML file "{local_path}": {str(err)}')
                if is_essential:
                    Log.error('Error: This XML file is essential. Abort! Please try again later!')
                    exit(2)
                else:
                    return None
        else:
            if is_essential:
                Log.error(f'Error: Can not find {local_path}. This XML file is essential. Please try again later!')
                exit(2)
            else:
                return None

    def final_mux(
        self,
        slideshow_path: str,
        webcams_path: str,
        webcams_rel_path: str,
        slideshow_width: int,
        slideshow_height: int,
        metadata: Metadata,
    ):

        webcam_is_empty = False
        if not self.skip_webcam_opt and not self.skip_webcam_freeze_detection_opt:
            Log.info(f'Try to detect freeze in {webcams_rel_path}...')
            with Timer() as t:
                webcam_is_empty = asyncio.run(self.ffmpeg.freeze_detect(webcams_path))

            Log.info(f'Detection of freeze finished and took: {formatSeconds(t.duration)}')
            if webcam_is_empty:
                Log.yellow('Webcam is empty, webcam will not be added to the final presentation')

        Log.info("Mux final slideshow")
        result_path = self.get_output_file_path(metadata)
        if os.path.isfile(result_path):
            Log.warning("Final Slideshow already exists. Abort!")
            exit(0)

        with Timer() as t:
            if self.skip_webcam_opt or webcam_is_empty:
                asyncio.run(
                    self.ffmpeg.add_audio_to_slideshow(
                        slideshow_path,
                        webcams_path,
                        result_path,
                    )
                )
            else:
                asyncio.run(
                    self.ffmpeg.add_webcam_to_slideshow(
                        slideshow_path,
                        webcams_path,
                        slideshow_width,
                        slideshow_height,
                        result_path,
                    )
                )

        Log.info(f'Mux final slideshow finished and took: {formatSeconds(t.duration)}')
        return result_path

    def add_deskshare_to_slideshow(
        self,
        slideshow_path: str,
        deskshare_path: str,
        deskshare_events: List[Deskshare],
        width: int,
        height: int,
        metadata: Metadata,
    ):
        if deskshare_path is None or len(deskshare_events) == 0:
            return slideshow_path

        presentation_path = PT.get_in_dir(self.tmp_dir, 'presentation.mp4')
        if os.path.isfile(presentation_path):
            Log.warning('Slideshow with deskshare does already exist! Skipping rendering!')
            return presentation_path

        Log.info('Resizing screen share...')
        resized_deskshare_path = PT.get_in_dir(self.tmp_dir, 'deskshare.mp4')
        if os.path.isfile(resized_deskshare_path):
            Log.warning('Resized screen share does already exist! Skipping rendering!')
        else:
            with Timer() as t:
                asyncio.run(self.ffmpeg.resize_deskshare(deskshare_path, resized_deskshare_path, width, height))
            Log.info(f'Resizing screen share finished and took: {formatSeconds(t.duration)}')

        Log.info('Start adding screen share to slideshow...')
        deskshare_txt_path = PT.get_in_dir(self.tmp_dir, 'deskshare.txt')
        with open(deskshare_txt_path, 'w', encoding="utf-8") as concat_file:
            for idx, event in enumerate(deskshare_events):
                if idx == 0 and event.start_timestamp > 0:
                    # Adding beginning
                    # duration = math.floor(10 * (event.start_timestamp) + 0.5) / 10
                    concat_file.write("file 'slideshow.mp4'\n")
                    concat_file.write("inpoint 0.0\n")
                    concat_file.write(f"outpoint {event.start_timestamp}\n")
                    # concat_file.write(f"duration {duration}\n")
                elif idx > 0:
                    # Adding part between deskshare
                    # duration = (
                    #     math.floor(10 * (event.start_timestamp - deskshare_events[idx - 1].stop_timestamp) + 0.5) / 10
                    # )
                    concat_file.write("file 'slideshow.mp4'\n")
                    concat_file.write(f"inpoint {deskshare_events[idx - 1].stop_timestamp}\n")
                    concat_file.write(f"outpoint {event.start_timestamp}\n")
                    # concat_file.write(f"duration {duration}\n")

                # Adding deskshare
                # duration = math.floor(10 * (event.stop_timestamp - event.start_timestamp) + 0.5) / 10
                concat_file.write("file 'deskshare.mp4'\n")
                concat_file.write(f"inpoint {event.start_timestamp}\n")
                concat_file.write(f"outpoint {event.stop_timestamp}\n")
                # concat_file.write(f"duration {duration}\n")

                if idx == (len(deskshare_events) - 1) and event.stop_timestamp < metadata.duration:
                    # Adding finish
                    # duration = math.floor(10 * (metadata.duration - event.stop_timestamp) + 0.5) / 10
                    concat_file.write("file 'slideshow.mp4'\n")
                    concat_file.write(f"inpoint {event.stop_timestamp}\n")
                    concat_file.write(f"outpoint {metadata.duration}\n")
                    # concat_file.write(f"duration {duration}\n")

        with Timer() as t:
            asyncio.run(self.ffmpeg.add_deskshare_to_slideshow(deskshare_txt_path, presentation_path))
        Log.info(f'Adding screen share to slideshow finished and took: {formatSeconds(t.duration)}')
        return presentation_path

    def create_slideshow(self, frames: Dict[float, Frame], width: int, height: int):
        Log.info('Start creating slideshow...')
        slideshow_path = PT.get_in_dir(self.tmp_dir, 'slideshow.mp4')
        if os.path.isfile(slideshow_path):
            Log.warning('Slideshow does already exist! Skipping rendering!')
            return slideshow_path

        slideshow_txt_path = PT.get_in_dir(self.frames_dir, 'slideshow.txt')
        with open(slideshow_txt_path, 'w', encoding="utf-8") as concat_file:
            timestamps = list(frames.keys())
            for idx in range(len(timestamps) - 1):
                duration = math.floor(10 * (timestamps[idx + 1] - timestamps[idx]) + 0.5) / 10
                concat_file.write(f"file '{frames[timestamps[idx]].capture_filename}'\n")
                concat_file.write(f"duration {duration}\n")

            # We use the second to last frame again, because the last frame is always empty.
            # concat_file.write(f"file {frames[timestamps[-2]].capture_filename}\n")

        with Timer() as t:
            asyncio.run(self.ffmpeg.create_slideshow(slideshow_txt_path, slideshow_path, width, height))
        Log.info(f'Creating slideshow finished and took: {formatSeconds(t.duration)}')
        return slideshow_path


def get_parser():
    """
    Creates a new argument parser.
    """
    parser = argparse.ArgumentParser(
        description=('Big Blue Button Downloader that downloads a BBB lesson as MP4 video')
    )

    parser.add_argument('URL', type=str, help='URL of a BBB lesson')

    parser.add_argument(
        '-sw',
        '--skip-webcam',
        action='store_true',
        help='Skip adding the webcam video as an overlay to the final video.'
        + ' This will reduce the time to generate the final video',
    )
    parser.add_argument(
        '-swfd',
        '--skip-webcam-freeze-detection',
        action='store_true',
        help='Skip detecting if the webcam video is completely empty.'
        + ' It is assumed the webcam recording is not empty. This will reduce the time to generate the final video',
    )

    parser.add_argument(
        '-sa',
        '--skip-annotations',
        action='store_true',
        help='Skip capturing the annotations of the professor. This will reduce the time to generate the final video',
    )

    parser.add_argument(
        '-sc',
        '--skip-cursor',
        action='store_true',
        help='Skip capturing the cursor of the professor. This will reduce the time to generate the final video',
    )

    parser.add_argument(
        '-bk',
        '--backup',
        action='store_true',
        help=(
            'Downloads all the content from the server and then stops. After using this option, you can run bbb-dl'
            + ' again to create the video based on the saved files'
        ),
    )
    parser.add_argument(
        '-kt',
        '--keep-tmp-files',
        action='store_true',
        help=(
            'Keep the temporary files after finish. In case of an error bbb-dl will reuse the already generated files'
        ),
    )

    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help=('Print more verbose debug information'),
    )

    parser.add_argument(
        '--ffmpeg-location',
        type=str,
        default=None,
        help=(
            'Optional path to the directory in that your installed ffmpeg executable is located'
            + ' (Use it if ffmpeg is not located in your system PATH)'
        ),
    )

    parser.add_argument(
        '-ncc',
        '--no-check-certificate',
        action='store_true',
        help=('Suppress HTTPS certificate validation'),
    )

    parser.add_argument(
        '--version',
        action='version',
        version='bbb-dl ' + __version__,
        help='Print program version and exit',
    )

    parser.add_argument(
        '--encoder',
        dest='encoder',
        type=str,
        default='libx264',
        help='Optional encoder to pass to ffmpeg (default libx264)',
    )
    parser.add_argument(
        '--audiocodec',
        dest='audiocodec',
        type=str,
        default='copy',
        help='Optional audiocodec to pass to ffmpeg (default copy the codec from the original source)',
    )

    parser.add_argument(
        '-f',
        '--filename',
        type=str,
        default=None,
        help='Optional output filename',
    )

    parser.add_argument(
        '-od',
        '--output-dir',
        type=str,
        default=None,
        help='Optional output directory for final video',
    )

    parser.add_argument(
        '-wd',
        '--working-dir',
        type=str,
        default=None,
        help='Optional output directory for all temporary directories/files',
    )

    parser.add_argument(
        '-mpc',
        '--max-parallel-chromes',
        type=int,
        default=10,
        help='Maximum number of chrome browser instances used to generate frames',
    )

    parser.add_argument(
        '-fw',
        '--force-width',
        type=int,
        default=None,
        help='Force width on final output. (e.g. 1280) This can reduce the time to generate the final video',
    )

    parser.add_argument(
        '-fh',
        '--force-height',
        type=int,
        default=None,
        help='Force height on final output. (e.g. 720) This can reduce the time to generate the final video',
    )

    return parser


# --- called at the program invocation: -------------------------------------
def main(args=None):
    just_fix_windows_console()
    parser = get_parser()
    args = parser.parse_args(args)

    with Timer() as final_t:
        BBBDL(
            args.URL,
            args.filename,
            args.output_dir,
            args.verbose,
            args.no_check_certificate,
            args.encoder,
            args.audiocodec,
            args.skip_webcam,
            args.skip_webcam_freeze_detection,
            args.skip_annotations,
            args.skip_cursor,
            args.keep_tmp_files,
            args.ffmpeg_location,
            args.working_dir,
            args.backup,
            args.max_parallel_chromes,
            args.force_width,
            args.force_height,
        ).run()
    Log.info(f'BBB-DL finished and took: {formatSeconds(final_t.duration)}')



import json
import os
import subprocess

from dataclasses import dataclass
from subprocess import CalledProcessError
from typing import List
from itertools import cycle

from ffmpeg import FFmpeg
from ffmpeg.utils import Progress

from bbb_dl.utils import PathTools as PT, Log


@dataclass
class VideoInfo:
    path: str
    duration: float
    width: int
    height: int


class FFMPEG:
    def __init__(self, verbose: bool, ffmpeg_location: str, encoder: str, audiocodec: str):
        self.verbose = verbose
        self.ffmpeg_path = 'ffmpeg'
        self.ffprobe_path = 'ffprobe'
        self.spinner = cycle('/|\\-')

        if ffmpeg_location is not None:
            found = False
            for check_name in ['ffmpeg', 'ffmpeg.exe']:
                check_path = PT.get_in_dir(ffmpeg_location, check_name)
                if os.path.isfile(check_path):
                    self.ffmpeg_path = check_path
                    found = True
            if not found:
                Log.error('Error: ffmpeg was not found in your specified --ffmpeg-location path')
                exit(-8)
            found = False
            for check_name in ['ffprobe', 'ffprobe.exe']:
                check_path = PT.get_in_dir(ffmpeg_location, check_name)
                if os.path.isfile(check_path):
                    self.ffprobe_path = check_path
                    found = True
            if not found:
                Log.error('Error: ffprobe was not found in your specified --ffmpeg-location path')
                exit(-9)

        self.encoder = encoder
        self.audiocodec = audiocodec
        self.stderr_log = []

    def on_error(self, code: int):
        if self.verbose:
            for line in self.stderr_log:
                print(line)
        else:
            Log.warning(
                'Please run bbb-dl again with the extra option --verbose to get the ffmpeg error message.'
                + ' Then add the log output to a new issue on https://github.com/C0D3D3V/bbb-dl/issues'
            )
        Log.error(f"Error: FFMpeg failed and returned error code {code}")
        exit(-10)

    def on_start(self, arguments: List[str]):
        self.stderr_log = []
        if self.verbose:
            Log.info(f"Running command: {' '.join(arguments)}")

    def on_progress(self, progress: Progress):
        print(
            f"\r\033[K{progress} {next(self.spinner)}",
            end='',
        )

    def on_completed(self):
        print()
        Log.info('Command finished')

    def on_log_stderr(self, line):
        if self.verbose:
            if line.find('bitrate=') == -1 and line.find('time=') == -1:
                self.stderr_log.append(line)

    def add_standard_handlers(self, ffmpeg_obj):
        ffmpeg_obj.on("start", self.on_start)
        ffmpeg_obj.on("error", self.on_error)
        ffmpeg_obj.on("progress", self.on_progress)
        ffmpeg_obj.on("stderr", self.on_log_stderr)
        ffmpeg_obj.on("completed", self.on_completed)

    def get_video_infos(self, video_path: str) -> VideoInfo:
        try:
            if self.verbose:
                Log.info(f'Checking video information of `{video_path}`')
            result = subprocess.run(
                [
                    self.ffprobe_path,
                    '-v',
                    'error',
                    '-select_streams',
                    "v:0",
                    "-show_entries",
                    "stream=width,height,duration",
                    "-of",
                    "json",
                    video_path,
                ],
                capture_output=True,
                encoding='utf-8',
                text=True,
                check=True,
            )
            streams = json.loads(result.stdout).get('streams', [])
            if len(streams) == 0:
                Log.warning(f"Error: No Stream found in {video_path}")
                return VideoInfo(video_path, None, 0, 0)
            stream = streams[0]
            return VideoInfo(video_path, stream.get('duration', None), stream.get('width', 0), stream.get('height', 0))
        except CalledProcessError as err:
            print(f"Error: {err}")
            exit(-10)

    async def freeze_detect(self, video_path: str) -> bool:
        """
        return true if video is 100% freezed
        """
        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            # .option("nostats")
            .input(video_path)
            .output(
                '-',
                vf='freezedetect=n=-60dB:d=2',
                map='0:v:0',
                f='null',
            )
        )

        freeze_starts = []
        freeze_ends = []

        @ffmpeg.on("stderr")
        def on_stderr(line):
            if line.find('lavfi.freezedetect.freeze_end') >= 0:
                end = float(line.rsplit('lavfi.freezedetect.freeze_end: ', 1)[1])
                freeze_ends.append(end)
            elif line.find('lavfi.freezedetect.freeze_start') >= 0:
                start = float(line.rsplit('lavfi.freezedetect.freeze_start: ', 1)[1])
                freeze_starts.append(start)

        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()
        if len(freeze_ends) == 0 and len(freeze_starts) == 1 and freeze_starts[0] <= 10:
            return True
        return False

    async def create_slideshow(self, concat_file_path: str, output_path: str, width: int, height: int):
        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            .input(
                concat_file_path,
                f='concat',
                # safe='0',
                # hwaccel="auto",  # In tests it was slower with hwaccel
            )
            .output(
                output_path,
                filter_complex=(
                    f'[0:v]fps=24,scale=w={width}:h={height}:force_original_aspect_ratio=decrease,'
                    + f'pad={width}:{height}:(ow-iw)/2:(oh-ih)/2[out]'
                ),
                map='[out]',
                strict='experimental',
                crf='22',
                pix_fmt='yuv420p',
                preset='ultrafast',
            )
        )
        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()

    async def resize_deskshare(self, deskshare_path: str, resized_deskshare_path: str, width: int, height: int):
        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            .input(
                deskshare_path,
                # hwaccel="auto", # Use encoder to activate hwaccel
            )
            .output(
                resized_deskshare_path,
                {
                    'c:v': self.encoder,
                    'c:a': self.audiocodec,
                },
                vf=(
                    f'scale=w={width}:h={height}:force_original_aspect_ratio=decrease,'
                    + f'pad={width}:{height}:(ow-iw)/2:(oh-ih)/2'
                ),
                preset='ultrafast',
            )
        )

        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()

    async def add_deskshare_to_slideshow(self, concat_file_path: str, output_path: str):
        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            .input(
                concat_file_path,
                f='concat',
                # safe='0',
                # hwaccel="auto",   # In tests it was slower with hwaccel
            )
            .output(
                output_path,
                {
                    'c:v': self.encoder,
                    'c:a': self.audiocodec,
                },
                strict='experimental',
                preset='ultrafast',
            )
        )
        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()

    def get_webcam_size(self, slideshow_width, slideshow_height):
        webcam_width = slideshow_width // 5
        webcam_height = webcam_width * 3 // 4

        if webcam_height > slideshow_height:
            webcam_height = slideshow_height

        if webcam_width % 2:
            webcam_width -= 1
        if webcam_height % 2:
            webcam_height -= 1

        return webcam_width, webcam_height

    async def add_webcam_to_slideshow(
        self,
        slideshow_path: str,
        webcams_path: str,
        slideshow_width: int,
        slideshow_height: int,
        result_path: str,
    ):
        webcam_width, webcam_height = self.get_webcam_size(slideshow_width, slideshow_height)

        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            .input(webcams_path)
            .input(slideshow_path)
            .output(
                result_path,
                {
                    'c:v': self.encoder,
                    'c:a': self.audiocodec,
                },
                filter_complex=(
                    f'[0:v]scale={webcam_width}:{webcam_height},setpts=PTS-STARTPTS,'
                    + 'format=rgba,colorchannelmixer=aa=0.8'
                    + '[ovrl];[1:v]fps=24,setpts=PTS-STARTPTS[bg];[bg][ovrl]overlay=W-w:H-h:shortest=1'
                ),
                strict='experimental',
                preset='ultrafast',
            )
        )
        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()

    async def add_audio_to_slideshow(self, slideshow_path: str, webcams_path: str, result_path: str):
        ffmpeg = (
            FFmpeg(self.ffmpeg_path)
            .option("hide_banner")
            .input(webcams_path)
            .input(slideshow_path)
            .output(
                result_path,
                {
                    'c:v': self.encoder,
                    'c:a': self.audiocodec,
                },
                map=['0:a', '1:v'],
                strict='experimental',
                preset='ultrafast',
                shortest=None,
            )
        )
        self.add_standard_handlers(ffmpeg)

        await ffmpeg.execute()



import collections
import functools
import html
import itertools
import math
import os
import re
import socket
import sys
import time
import unicodedata

from contextlib import closing
from http.server import SimpleHTTPRequestHandler
from pathlib import Path


class QuietRequestHandler(SimpleHTTPRequestHandler):
    def log_message(self, *args):
        pass


def get_free_port():
    """
    Ask the system for a free port.
    In case of error return error message.
    :return: {Tuple}
    """
    port = None
    error = {}
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as soc:
        try:
            soc.bind(('127.0.0.1', 0))
            sock_name = soc.getsockname()
            if isinstance(sock_name, tuple) and len(sock_name) == 2:
                port = sock_name[1]
        except socket.error as e:
            error = {'errno': e.errno, 'msg': str(e)}

        return port, error


def check_verbose() -> bool:
    """Return if the verbose mode is active"""
    return '-v' in sys.argv or '--verbose' in sys.argv


def check_debug() -> bool:
    """Return if the debugger is currently active"""
    return 'pydevd' in sys.modules or (hasattr(sys, 'gettrace') and sys.gettrace() is not None)


_timetuple = collections.namedtuple('Time', ('hours', 'minutes', 'seconds', 'milliseconds'))


def float_or_none(v, scale=1, invscale=1, default=None):
    if v is None:
        return default
    try:
        return float(v) * invscale / scale
    except (ValueError, TypeError):
        return default


def format_decimal_suffix(num, fmt='%d%s', *, factor=1000):
    """Formats numbers with decimal sufixes like K, M, etc"""
    num, factor = float_or_none(num), float(factor)
    if num is None or num < 0:
        return None
    POSSIBLE_SUFFIXES = 'kMGTPEZY'
    exponent = 0 if num == 0 else min(int(math.log(num, factor)), len(POSSIBLE_SUFFIXES))
    suffix = ['', *POSSIBLE_SUFFIXES][exponent]
    if factor == 1024:
        suffix = {'k': 'Ki', '': ''}.get(suffix, f'{suffix}i')
    converted = num / (factor**exponent)
    return fmt % (converted, suffix)


def format_bytes(bytes):
    return format_decimal_suffix(bytes, '%.2f%sB', factor=1024) or 'N/A'


def append_get_idx(list_obj, item):
    idx = len(list_obj)
    list_obj.append(item)
    return idx


def timetuple_from_msec(msec):
    secs, msec = divmod(msec, 1000)
    mins, secs = divmod(secs, 60)
    hrs, mins = divmod(mins, 60)
    return _timetuple(hrs, mins, secs, msec)


def formatSeconds(secs, delim=':', msec=False):
    time = timetuple_from_msec(secs * 1000)
    if time.hours:
        ret = '%d%s%02d%s%02d' % (time.hours, delim, time.minutes, delim, time.seconds)
    elif time.minutes:
        ret = '%d%s%02d' % (time.minutes, delim, time.seconds)
    else:
        ret = '%d' % time.seconds
    return '%s.%03d' % (ret, time.milliseconds) if msec else ret


def xpath_text(node, xpath):
    n = node.find(xpath)
    if n is None:
        return n
    if n.text is None:
        return None
    return n.text


def xpath_with_ns(path, ns_map):
    components = [c.split(':') for c in path.split('/')]
    replaced = []
    for c in components:
        if len(c) == 1:
            replaced.append(c[0])
        else:
            ns, tag = c
            replaced.append('{%s}%s' % (ns_map[ns], tag))
    return '/'.join(replaced)


_s = functools.partial(
    xpath_with_ns,
    ns_map={'svg': 'http://www.w3.org/2000/svg'},
)
_x = functools.partial(
    xpath_with_ns,
    ns_map={
        # 'xmlns': 'http://www.w3.org/2000/svg',
        'xml': 'http://www.w3.org/XML/1998/namespace',
        'ttml': 'http://www.w3.org/ns/ttml',
        'tts': 'http://www.w3.org/ns/ttml#styling',
        'xlink': 'http://www.w3.org/1999/xlink',
    },
)


RESET_SEQ = '\033[0m'
COLOR_SEQ = '\033[1;%dm'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)


class Log:
    """
    Logs a given string to output with colors
    :param logString: the string that should be logged

    The string functions returns the strings that would be logged.
    """

    @staticmethod
    def info_str(logString: str):
        return COLOR_SEQ % WHITE + logString + RESET_SEQ

    @staticmethod
    def success_str(logString: str):
        return COLOR_SEQ % GREEN + logString + RESET_SEQ

    @staticmethod
    def warning_str(logString: str):
        return COLOR_SEQ % YELLOW + logString + RESET_SEQ

    @staticmethod
    def yellow_str(logString: str):
        return COLOR_SEQ % YELLOW + logString + RESET_SEQ

    @staticmethod
    def error_str(logString: str):
        return COLOR_SEQ % RED + logString + RESET_SEQ

    @staticmethod
    def debug_str(logString: str):
        return COLOR_SEQ % CYAN + logString + RESET_SEQ

    @staticmethod
    def blue_str(logString: str):
        return COLOR_SEQ % BLUE + logString + RESET_SEQ

    @staticmethod
    def magenta_str(logString: str):
        return COLOR_SEQ % MAGENTA + logString + RESET_SEQ

    @staticmethod
    def info(logString: str):
        print(Log.info_str(logString))

    @staticmethod
    def success(logString: str):
        print(Log.success_str(logString))

    @staticmethod
    def warning(logString: str):
        print(Log.warning_str(logString))

    @staticmethod
    def yellow(logString: str):
        print(Log.yellow_str(logString))

    @staticmethod
    def error(logString: str):
        print(Log.error_str(logString))

    @staticmethod
    def debug(logString: str):
        print(Log.debug_str(logString))

    @staticmethod
    def blue(logString: str):
        print(Log.blue_str(logString))

    @staticmethod
    def magenta(logString: str):
        print(Log.magenta_str(logString))


class Timer(object):
    '''
    Timing Context Manager
    Can be used for future speed comparisons, like this:

    with Timer() as t:
        Do.stuff()
    print(f'Do.stuff() took:\t {t.duration:.3f} \tseconds.')
    '''

    def __enter__(self):
        self.start = time.perf_counter_ns()
        return self

    def __exit__(self, *args):
        end = time.perf_counter_ns()
        self.duration = (end - self.start) * 10**-9  # 1 nano-sec = 10^-9 sec


NO_DEFAULT = object()

# needed for sanitizing filenames in restricted mode
ACCENT_CHARS = dict(
    zip(
        'ÂÃÄÀÁÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖŐØŒÙÚÛÜŰÝÞßàáâãäåæçèéêëìíîïðñòóôõöőøœùúûüűýþÿ',
        itertools.chain(
            'AAAAAA',
            ['AE'],
            'CEEEEIIIIDNOOOOOOO',
            ['OE'],
            'UUUUUY',
            ['TH', 'ss'],
            'aaaaaa',
            ['ae'],
            'ceeeeiiiionooooooo',
            ['oe'],
            'uuuuuy',
            ['th'],
            'y',
        ),
    )
)


class PathTools:
    """A set of methods to create correct paths."""

    restricted_filenames = False

    @staticmethod
    def to_valid_name(name: str) -> str:
        """Filtering invalid characters in filenames and paths.

        Args:
            name (str): The string that will go through the filtering

        Returns:
            str: The filtered string, that can be used as a filename.
        """

        if name is None:
            return None

        name = html.unescape(name)

        name = name.replace('\n', ' ')
        name = name.replace('\r', ' ')
        name = name.replace('\t', ' ')
        name = name.replace('\xad', '')
        while '  ' in name:
            name = name.replace('  ', ' ')
        name = PathTools.sanitize_filename(name, PathTools.restricted_filenames)
        name = name.strip('. ')
        name = name.strip()

        return name

    @staticmethod
    def sanitize_filename(s, restricted=False, is_id=NO_DEFAULT):
        """Sanitizes a string so it could be used as part of a filename.
        @param restricted   Use a stricter subset of allowed characters
        @param is_id        Whether this is an ID that should be kept unchanged if possible.
                            If unset, yt-dlp's new sanitization rules are in effect
        """
        if s == '':
            return ''

        def replace_insane(char):
            if restricted and char in ACCENT_CHARS:
                return ACCENT_CHARS[char]
            elif not restricted and char == '\n':
                return '\0 '
            elif is_id is NO_DEFAULT and not restricted and char in '"*:<>?|/\\':
                # Replace with their full-width unicode counterparts
                return {'/': '\u29F8', '\\': '\u29f9'}.get(char, chr(ord(char) + 0xFEE0))
            elif char == '?' or ord(char) < 32 or ord(char) == 127:
                return ''
            elif char == '"':
                return '' if restricted else '\''
            elif char == ':':
                return '\0_\0-' if restricted else '\0 \0-'
            elif char in '\\/|*<>':
                return '\0_'
            if restricted and (char in '!&\'()[]{}$;`^,#' or char.isspace() or ord(char) > 127):
                return '\0_'
            return char

        if restricted and is_id is NO_DEFAULT:
            s = unicodedata.normalize('NFKC', s)
        s = re.sub(r'[0-9]+(?::[0-9]+)+', lambda m: m.group(0).replace(':', '_'), s)  # Handle timestamps
        result = ''.join(map(replace_insane, s))
        if is_id is NO_DEFAULT:
            result = re.sub(r'(\0.)(?:(?=\1)..)+', r'\1', result)  # Remove repeated substitute chars
            STRIP_RE = r'(?:\0.|[ _-])*'
            result = re.sub(f'^\0.{STRIP_RE}|{STRIP_RE}\0.$', '', result)  # Remove substitute chars from start/end
        result = result.replace('\0', '') or '_'

        if not is_id:
            while '__' in result:
                result = result.replace('__', '_')
            result = result.strip('_')
            # Common case of "Foreign band name - English song title"
            if restricted and result.startswith('-_'):
                result = result[2:]
            if result.startswith('-'):
                result = '_' + result[len('-') :]
            result = result.lstrip('.')
            if not result:
                result = '_'
        return result

    @staticmethod
    def remove_start(s, start):
        return s[len(start) :] if s is not None and s.startswith(start) else s

    @staticmethod
    def sanitize_path(path: str):
        """
        @param path: A path to sanitize.
        @return: A path where every part was sanitized using to_valid_name.
        """
        drive_or_unc, _ = os.path.splitdrive(path)
        norm_path = os.path.normpath(PathTools.remove_start(path, drive_or_unc)).split(os.path.sep)
        if drive_or_unc:
            norm_path.pop(0)

        sanitized_path = [
            path_part if path_part in ['.', '..'] else PathTools.to_valid_name(path_part) for path_part in norm_path
        ]

        if drive_or_unc:
            sanitized_path.insert(0, drive_or_unc + os.path.sep)
        return os.path.join(*sanitized_path)

    @staticmethod
    def get_abs_path(path: str):
        return str(Path(path).resolve())

    @staticmethod
    def get_in_dir(path: str, filename: str):
        return str(Path(path) / filename)

    @staticmethod
    def make_base_dir(path_to_file: str):
        Path(path_to_file).parent.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def make_dirs(path_to_dir: str):
        Path(path_to_dir).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def get_user_data_directory():
        """Returns a platform-specific root directory for user application data."""
        if os.name == "nt":
            appdata = os.getenv("LOCALAPPDATA")
            if appdata:
                return appdata
            appdata = os.getenv("APPDATA")
            if appdata:
                return appdata
            return None
        # On non-windows, use XDG_DATA_HOME if set, else default to ~/.config.
        xdg_config_home = os.getenv("XDG_DATA_HOME")
        if xdg_config_home:
            return xdg_config_home
        return os.path.join(os.path.expanduser("~"), ".local/share")

    @staticmethod
    def get_project_data_directory():
        """
        Returns an Path object to the project config directory
        """
        data_dir = Path(PathTools.get_user_data_directory()) / "bbb-dl"
        if not data_dir.is_dir():
            data_dir.mkdir(parents=True, exist_ok=True)
        return str(data_dir)
